import sys
import click
import logging
import io

from . import __version__, number_files, _log
from .exceptions import FnumException


class _ClickFormatter(logging.Formatter):
    def format(self, record):
        log_str = super().format(record)
        click.echo(log_str)
        return log_str


@click.command(
    help="""
Renames files in a directory using sequential integers.\n
Suffixes is a comma separated list of file extensions to rename (eg. .jpg,.gif).\n
Dirpath is a directory to rename files in.
""",
    context_settings={
        "help_option_names": ["-h", "--help"],
    },
)
@click.argument("suffixes", nargs=1)
@click.argument("dirpath", nargs=1)
@click.version_option(version=__version__)
@click.option(
    "--write-max/--no-write-max",
    default=False,
    help="""
Write a file named fnum.max.txt containing the highest number used in a filename.\n
This is useful for being able to list or paginate numbered files when listing the files is not otherwise possible (such as on a static website).
    """,
)
@click.option(
    "--write-metadata/--no-write-metadata",
    default=False,
    help="""
Writes a file named fnum.metadata.yaml containing data about what was changed.\n
Max is the highest number used in a filename.\n
Order contains a list of filenames in the order they were added. Newly added files are added to the end of the list when the command is run multiple times.\n
Originals maps the original filenames to what they were renamed to.
    """,
)
@click.option(
    "--include-imeta/--no-include-imeta",
    default=False,
    help="""
Also rename image metadata files generated by imeta.
    """,
)
def cli(**kwargs):
    dirpath = kwargs["dirpath"]
    if "/" in kwargs["suffixes"]:
        click.echo("Suffixes contains a '/', did you mean ','?", err=True)
    suffixes = kwargs["suffixes"].split(",")

    _log.setLevel(logging.INFO)
    handler = logging.StreamHandler(io.StringIO())
    handler.setFormatter(_ClickFormatter())
    _log.addHandler(handler)

    try:
        try:
            number_files(
                dirpath=dirpath,
                suffixes=suffixes,
                write_metadata=kwargs["write_metadata"],
                write_max=kwargs["write_max"],
                include_imeta=kwargs["include_imeta"],
            )
        finally:
            _log.removeHandler(handler)
    except (FnumException, FileNotFoundError) as e:
        click.echo(str(e), err=True)
        sys.exit(1)
